'''
    author: rkopeinig
    script: Preprocessing
    description: Bunch of pre-processing algorithms.
                 See Proba-V Time-Series Analysis Toolbox Poster.
    version: 0.1
'''

import numpy as np
import pandas as pd
from math import sqrt, pi, log
from numpy import fft
from sklearn.metrics import mean_squared_error
from scipy import interpolate
from scipy.signal import savgol_filter


def fourierExtrapolation(series, n_predict=100):
    """Fourier Extrapolation on given Time-Series.

        :param series: One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param n_predict: Number of predictions
        :type n_predict: integer

        :return: Fourier Extrapolation results of given Time-Series
        :rtype: pandas.core.series.Series
    """

    try:
        n = series.size
        n_harm = 5
        t = np.arange(0, n)
        p = np.polyfit(t, series, 1)
        x_notrend = series - p[0] * t
        x_freqdom = fft.fft(x_notrend)
        f = fft.fftfreq(n)
        indexes = list(range(n))
        indexes.sort(key=lambda i: np.absolute(f[i]))
        t = np.arange(0, n + n_predict)
        restored_sig = np.zeros(t.size)
        for i in indexes[:1 + n_harm * 2]:
            ampli = np.absolute(x_freqdom[i]) / n
            phase = np.angle(x_freqdom[i])
            restored_sig += ampli * np.cos(2 * np.pi * f[i] * t + phase)
        return pd.Series(restored_sig + p[0] * t)
    except Exception as e:
        print('Error = %s' % str(e))


def rolling_min(series, window):
    """Provides rolling window calculations of Time-Series

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param window: Size of moving window. Number of observations to calculate statistics
        :type window: integer

        :return: Rolling minimum for given Time Series
        :rtype: pandas.core.series.Series
    """
    try:
        return series.rolling(window).min()
    except Exception as e:
        print('Error = %s' % str(e))

def rolling_mean(series, window):
    """Provides rolling window calculations of Time-Series

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param window: Size of moving window. Number of observations to calculate statistics
        :type window: integer

        :return: Rolling mean for given Time Series
        :rtype: pandas.core.series.Series
    """
    try:
        return series.rolling(window).mean()
    except Exception as e:
        print('Error = %s' % str(e))

def rolling_median(series, window):
    """Provides rolling window calculations of Time-Series

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param window: Size of moving window. Number of observations to calculate statistics
        :type window: integer

        :return: Rolling median for given Time Series
        :rtype: pandas.core.series.Series
    """
    try:
        return series.rolling(window).median()
    except Exception as e:
        print('Error = %s' % str(e))


def rolling_std(series, window):
    """Provides rolling window calculations of Time-Series

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param window: Size of moving window. Number of observations to calculate statistics
        :type window: integer

        :return: Rolling standard deviation for given Time Series
        :rtype: pandas.core.series.Series
    """
    try:
        return series.rolling(window).std()
    except Exception as e:
        print('Error = %s' % str(e))


def rolling_max(series, window):
    """Provides rolling window calculations of Time-Series

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param window: Size of moving window. Number of observations to calculate statistics
        :type window: integer

        :return: Rolling maximum for given Time Series
        :rtype: pandas.core.series.Series
    """
    try:
        return series.rolling(window).max()
    except Exception as e:
        print('Error = %s' % str(e))


def seasonal_filtering(series, filter='M'):
    """ Seasonal filtering

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param filter: String which indicates the method to be used
        :type series: str

        :return: Pandas Series of smoothed Time-Series using Scipy's spline interpolation methods.
        :rtype: pandas.core.series.Series

    """
    try:
        return series.resample(filter, how='mean')
    except Exception as e:
        print('Error = %s' % str(e))


def outlier_analysis_using_filtered_median(series, threshold=3):
    """Returns outliers using median filtering (non-linear) technique.

        :param series: One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param threshold: Median-based threshold
        :type threshold: integer

        :return: Series of outliers of Time-Series
        :rtype: pandas.core.series.Series

    """
    try:
        signal = series.copy()
        difference = np.abs(signal - np.median(signal))
        median_difference = np.median(difference)
        if median_difference == 0:
            s = 0
        else:
            s = difference / float(median_difference)
        mask = s > threshold
        signal[mask] = np.median(signal)
        return series[(np.where(signal.values != series.values)[0])]
    except Exception as e:
        print('Error = %s' % str(e))


def savitz_golay(series, window=10, polyorder=2):
    """Returns smoothed Time-Series using Savitzky Golay.

                :param series: One-dimensional ndarray with axis labels including Time-Series.
                :type series: pandas.core.series.Series

                :param window: Window size has to be an odd number
                :type window: integer

                :param polyorder: Polyorder has to be smaller than window param
                :type window: integer

                :return: Series of smoothed Time-Series
                :rtype: pandas.core.series.Series

    """
    try:
        if type(series) == pd.Series:
            if window % 2 == 0:
                res = savgol_filter(series.values, window - 1, polyorder)
            else:
                res = savgol_filter(series.values, window, polyorder)
            return pd.Series(res, index=series.index)

        if type(series) == pd.DataFrame:
            if window % 2 == 0:
                return series.apply(lambda x: savgol_filter(x.values, window + 1, polyorder))
            else:
                return series.apply(lambda x: savgol_filter(x.values, window, polyorder))
    except Exception as e:
        print('Error = %s' % str(e))


def temporal_smoothing_pandas(series, method='cubic'):
    """ Temporal Smoothing using Cubic and Spline

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param method: Smoothing Method either cubic or spline. Cubic is set as default
        :type method: string

        :return: Pandas Series of smoothed Time-Series using cubic and spline interpolation methods.
        :rtype: pandas.core.series.Series

    """
    try:
        if method=='cubic':
            return series.interpolate(method='cubic')
        if method=='spline':
            return series.interpolate(method='spline', order=3, s=0)
    except Exception as e:
        print('Error = %s' % str(e))

def temporal_smoothing_scipy(series):
    """ Temporal Smoothing using Spline

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :return: Pandas Series of smoothed Time-Series using Scipy's spline interpolation methods.
        :rtype: pandas.core.series.Series

    """
    try:
        x = range(1, len(series) + 1)
        tck = interpolate.splrep(x, series.values)
        return pd.Series(interpolate.splev(x, tck),index=series.index)
    except Exception as e:
        print('Error = %s' % str(e))