'''
    author: rkopeinig
    script: Basic Statistics and Trend Analysis
    description: Bunch of basic statistics and trend analysis algorithms.
                 See Proba-V Time-Series Analysis Toolbox Poster.
    version: 0.1
'''

import numpy as np
import pandas as pd
from scipy.stats import norm
from statsmodels.tsa.seasonal import seasonal_decompose
from statsmodels.regression.linear_model import OLS
from statsmodels.tsa.stattools import adfuller



def adfuller_test(series):
    """ Augmented Dickey-Fuller test can be used to test for a unit root in a univariate process in the presence of serial correlation.

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :return: Pandas DataFrame of result dictionary.
        :rtype: pandas.DataFrame

    """
    try:
        log_series = series.apply(lambda x: np.log(x))
        log_series_diff=log_series-log_series.shift()
        log_series_diff.dropna(inplace=True)
        raw = adfuller(log_series_diff, autolag='AIC', regression='c')[:5]
        data = [v for v in raw[:4]]
        data.extend([raw[4]['1%'], raw[4]['5%'], raw[4]['10%']])
        return pd.DataFrame({'Value':data, 'Label': ['Test Statistics', 'p-value','#Lags used', 'Number of Observations Used', 'Critical value for 1%', 'Critical value for 5%', 'Critical value for 10%']})
    except Exception as e:
        print('Error = %s' % str(e))

def de_trender(series):
    """ Perform a de-trending on a given Time-Series using log.

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :return: Pandas Series of detrended Time-Series.
        :rtype: pandas.core.series.Series

    """
    try:
        log_series = series.apply(lambda x: np.log(x))
        log_series_diff = log_series - log_series.shift()
        return log_series_diff
    except Exception as e:
        print('Error = %s' % str(e))


def mann_kendall(series, alpha=0.05):
    """ Kendall's tau coefficient to measure the ordinal association between two measured quantities

        :param series: One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param alpha: Significance level
        :type alpha: integer or float

        :return param trend: Returns if trend is increasing, decreasing or there is no trend
        :rtype: string

        :return param h: Returns if trend is present or absence
        :rtype: boolean

        :return param p: P value of the significance test
        :rtype: float

        :return param z: Normalized test statistics
        :rtype: float

    """
    try:
        n = len(series)
        s = 0
        for k in range(n - 1):
            for j in range(k + 1, n):
                s += np.sign(series[j] - series[k])

        unique_x = np.unique(series)
        g = len(unique_x)

        if n == g:  # no tie
            var_s = (n * (n - 1) * (2 * n + 5)) / 18
        else:  # some ties in data
            tp = np.zeros(unique_x.shape)
            for i in range(len(unique_x)):
                tp[i] = sum(unique_x[i] == series)
            var_s = (n * (n - 1) * (2 * n + 5) + np.sum(tp * (tp - 1) * (2 * tp + 5))) / 18

        if s > 0:
            z = (s - 1) / np.sqrt(var_s)
        elif s == 0:
            z = 0
        elif s < 0:
            z = (s + 1) / np.sqrt(var_s)

        p = 2 * (1 - norm.cdf(abs(z)))  # two tail test
        h = abs(z) > norm.ppf(1 - alpha / 2)

        if (z < 0) and h:
            trend = 'decreasing'
        elif (z > 0) and h:
            trend = 'increasing'
        else:
            trend = 'no trend'
        return trend, h, p, z
    except Exception as e:
        print('Error = %s' % str(e))

def seasonal_decomposition(series, model, freq):
    """ Seasonal Decompositon
        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param freq: Frequency of the series. Must be used if x is not a pandas object.
        :type freq: integer

        :return param df: DataFrame of seasonal decomposition results including seasonal, trend, and resid attributes.
        :rtype param df: pandas.DataFrame

    """
    try:
        log_series = series.apply(lambda x: np.log(x))
        log_series.interpolate(inplace=True)
        return seasonal_decompose(log_series, model=model, freq=freq)
    except Exception as e:
        print('Error = %s' % str(e))

def _datacheck_peakdetect(x_axis, y_axis):
    """Helper function for peak detection. Checks if length of both axis is the same and converts them into numpy arrays

                    :param x_axis: X axis for the index.
                    :type x_axis: numpy.ndarray

                    :param y_axis: Y axis for the actual data.
                    :type y_axis: numpy.ndarray

                    :return x_axis: X axis as numpy array
                    :rtype x_axis: numpy.array

                    :return y_axis: Y axis as numpy array
                    :rtype y_axis: numpy.array

        """
    try:
        if x_axis is None:
            x_axis = range(len(y_axis))

        if len(y_axis) != len(x_axis):
            raise (ValueError,
                   'Input vectors y_axis and x_axis must have same length')

        # needs to be a numpy array
        y_axis = np.array(y_axis)
        x_axis = np.array(x_axis)
        return x_axis, y_axis
    except Exception as e:
        print('Error = %s' % str(e))


def peakdetect(series, lookahead=10, delta=0):
    """Returns low and high peaks for given Time-Series.

                :param series: One-dimensional ndarray with axis labels including Time-Series.
                :type series: pandas.core.series.Series

                :param lookahead: To look ahead from a peak candidate to determine if it is the actual peak
                :type lookahead: integer

                :param delta: Specifies min difference between a peak and following observation before considering it as a peak
                :type delta: integer

                :return _min_: Series of low peaks
                :rtype _min_: pandas.core.series.Series

                :return _max_: Series of high peaks
                :rtype _max_: pandas.core.series.Series

    """
    try:
        max_peaks = []
        min_peaks = []
        dump = []
        dataframe = pd.DataFrame({'input': series}, index=series.index)
        y_axis = series.values
        series = series.reset_index()
        x_axis = series.index.values

        x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
        length = len(y_axis)

        if lookahead < 1:
            raise ValueError("Lookahead must be '1' or above in value")
        if not (np.isscalar(delta) and delta >= 0):
            raise ValueError("delta must be a positive number")

        mn, mx = np.Inf, -np.Inf

        for index, (x, y) in enumerate(zip(x_axis[:-lookahead],
                                           y_axis[:-lookahead])):
            if y > mx:
                mx = y
                mxpos = x
            if y < mn:
                mn = y
                mnpos = x

            if y < mx - delta and mx != np.Inf:
                if y_axis[index:index + lookahead].max() < mx:
                    max_peaks.append([mxpos, mx])
                    dump.append(True)
                    mx = np.Inf
                    mn = np.Inf
                    if index + lookahead >= length:
                        break
                    continue
            if y > mn + delta and mn != -np.Inf:
                if y_axis[index:index + lookahead].min() > mn:
                    min_peaks.append([mnpos, mn])
                    dump.append(False)
                    mn = -np.Inf
                    mx = -np.Inf
                    if index + lookahead >= length:
                        break
        try:
            if dump[0]:
                max_peaks.pop(0)
            else:
                min_peaks.pop(0)
            del dump
        except IndexError:
            pass

        xm = [p[0] for p in max_peaks]
        ym = [p[1] for p in max_peaks]
        xn = [p[0] for p in min_peaks]
        yn = [p[1] for p in min_peaks]

        _maxdf = pd.DataFrame(ym, index=xm)
        _max_input = dataframe.reset_index()
        _max_input['data'] = _maxdf
        del _max_input['input']
        _max_input = _max_input[np.isfinite(_max_input['data'])]
        _max_ = _max_input.set_index('index')

        _mindf = pd.DataFrame(yn, index=xn)
        _min_input = dataframe.reset_index()
        _min_input['data'] = _mindf
        del _min_input['input']
        _min_input = _min_input[np.isfinite(_min_input['data'])]
        _min_ = _min_input.set_index('index')

        return [_min_, _max_]
    except Exception as e:
        print('Error = %s' % str(e))


def descriptive_statistics(df):
    """ Descriptive Statistic

        :param df: Two-dimensional size-mutable ndarray with axis labels including Time-Series.
        :type df: pandas.core.series.Series

        :return param df: DataFrame of descriptive statistics about given Pandas DataFrame.
        :rtype param df: pandas.DataFrame

    """
    try:
        grouped_data = df.groupby(['data'])
        return grouped_data['data'].describe().unstack()
    except Exception as e:
        print('Error = %s' % str(e))

def pivot_table(df, columns='year', index='month'):
    """ Pivot Table

        :param df: Two-dimensional size-mutable ndarray with axis labels including Time-Series.
        :type df: pandas.core.series.Series

        :return: DataFrame Pivot Table with yearly, monthly or daily attributes.
        :rtype: pandas.DataFrame.pivot

    """
    try:
        df['year'] = df.index.year
        df['month'] = df.index.month
        df['day'] = df.index.day
        return pd.pivot_table(df, values='data', columns=columns, index=index)
    except Exception as e:
        print('Error = %s' % str(e))